﻿Telepathy = LibStub("AceAddon-3.0"):NewAddon("Telepathy", "AceConsole-3.0", "AceEvent-3.0", "AceComm-3.0", "AceTimer-3.0", "AceSerializer-3.0" );

--local function cache
local Time = Time
local ipairs = ipairs
local pairs = pairs
local tinsert = tinsert

--debug tracing switch
local teleDebug = false

--
local ButtonCache = 
	{ 
		Foreign = {}, 
		Local = {}
	}
--[[ LinkedMembers is the table of party/raid members who have published their CDs
	Entry:
		{
			Name,
			Abilities = { {info, hidden} },
			PanelButtons = { buttonrefs },
			Panel = wrappanel
		}
]]
local LinkedMembers = {} 

--[[ PublishedAbilities is the table of your abilities/items that you have published, containing
	details about the ability as well as a reference to the actual ability button]]
local PublishedAbilities = {}

-- button dimensions
local borderPadding = 10

-- opacity settings
local opacityCD = 0.7
local opacityAvailable = 1

-- comm prefixes
local comPublish = "TelepathyPub"
local comCD = "TelepathyCD"
local comRequest = "TelepathyReq"
local comRefresh = "TelepathyRefresh"

-- request sound files
local sounds = {
	["--None--"] = "",
	["Wisp"] = "Sound\\Event Sounds\\Wisp\\WispYes1.wav",
	["Ding"] = "Sound\\interface\\AuctionWindowOpen.wav",
	["Magic Wand"] = "Sound\\Item\\UseSounds\\iMagicWand1.wav",
	["Bull Whip"] = "Sound\\Item\\Weapons\\Whip\\BullWhipHit1.wav",
	["PVP Flag Return"] = "Sound\\Spells\\PVPFlagReturned.wav",
	["PVP Enter Queue"] = "Sound\\Spells\\PVPEnterQueue.wav",
	["Bell"] = "Sound\\Spells\\ShaysBell.wav",
	["Spell Break"] = "Sound\\Spells\\SpellBreakTarget.wav",
	["Throw Net"] = "Sound\\Spells\\ThrowNet.wav",
	["Bonk 1"] = "Sound\\Spells\\Bonk1.wav",
	["Bonk 2"] = "Sound\\Spells\\Bonk2.wav",
	["Bonk 3"] = "Sound\\Spells\\Bonk3.wav",
	["Illidan Blades"] = "Sound\\Spells\\IllidanGetBlades.wav",
	["Rein Deer"] = "Sound\\Spells\\SummonReindeer.wav"
	}

-- current comm channel (to change when party/raid changes)
local currentChannel = "PARTY"

-- the spell book for chosing which spells you are interested in
local Telepathy_SpellBook = nil

--colors for player name labels
local Colors = {
		["Class"]={
					["HUNTER"] = { 0.67, 0.83, 0.45, 1 },
					["WARLOCK"] = { 0.58, 0.51, 0.79, 1 },
					["PRIEST"] = { 1.0, 1.0, 1.0, 1 },
					["PALADIN"] = { 0.96, 0.55, 0.73, 1 },
					["MAGE"] = { 0.41, 0.8, 0.94, 1 },
					["ROGUE"] = { 1.0, 0.96, 0.41, 1 },
					["DRUID"] = { 1.0, 0.49, 0.04, 1 },
					["SHAMAN"] = { 0.14, 0.35, 1.0, 1 },
					["WARRIOR"] = { 0.78, 0.61, 0.43, 1 },
					["DEATHKNIGHT"] = { 0.77, 0.12, 0.23, 1 },
					["UNKNOWN"] = { 0.77, 0.77, 0.77, 1 }
					}
}

--options
local options = {
    name = "Telepathy",
    handler = Telepathy,
    type = "group",
    args = {
		SpellBook = {
			order = 0,
			name = "Spell Book",
			type = "execute",
			desc = "Open/Close the Telepathy Spellbook",
			func = function() Telepathy:CreateSpellBook() end,
		},
		Layout = {
			name = "Layout",
			type = "group",
			guiInline = true,
			args = {
				help = {
					order = 0,
					type = "description",
					name = "Modify the button size and padding options for the combat window",
				},
				mybuttonsize = {
					order = 1,
					type = "range",			
					min = 16,
					max = 54,
					step = 1,
					bigStep = 1,
					name = "Button Size",
					desc = "The size of your own ability buttons in the combat window and spell book",            
					get = "GetButtonSizeLocal",
					set = "SetButtonSizeLocal",
					width = "double"
				},
				foreignbuttonsize = {
					order = 2,
					type = "range",			
					min = 16,
					max = 54,
					step = 1,
					bigStep = 1,
					name = "Button Size",
					desc = "The size of foreign ability buttons in the combat window and spell book",            
					get = "GetButtonSizeForeign",
					set = "SetButtonSizeForeign",
					width = "double"
				},
				buttonpadding = {
					type = "range",
					min = 0,
					max = 20,
					step = 1,
					bigStep = 1,
					name = "Button Padding",
					desc = "The padding space between ability buttons in the combat window and spell book",
					get = "GetButtonPadding",
					set = "SetButtonPadding",
					width = "double"
				},
			}		
		},
		Opacity = {
			name = "Opacity",
			type = "group",
			guiInline = true,
			args = {
				help = {
					order = 0,
					type = "description",
					name = "Opacity affects the transparency of the combat window buttons, depending on whether or not the cooldown is available",
				},
				oncdopacity = {
					order = 1,
					type = "range",
					min = 0,
					max = 1,
					step = 0.05,
					bigStep = 0.05,
					name = "On Cooldown Opacity",
					desc = "The opacity level for abilities which are on cooldown",
					get = "GetOpacityCD",
					set = "SetOpacityCD",			
					width = "double"
				},
				offcdopacity = {
					order = 2,
					type = "range",
					min = 0,
					max = 1,
					step = 0.05,
					bigStep = 0.05,
					name = "Available Opacity",
					desc = "The opacity level for abilities which are available",
					get = "GetOpacityAvailable",
					set = "SetOpacityAvailable",			
					width = "double"
				},
			}
		},
		nametrim = {
            type = "range",
			min = 0,
			max = 8,
			step = 1,
			bigStep = 1,
            name = "Name Trim Length",
            desc = "The abbreviated player name label length to show for other player's abilities. Set to ZERO for no trimming. Changes to this may require a interface reload.",
            get = "GetNameTrimLength",
            set = "SetNameTrimLength",			
			width = "double"
        },
		RequestSound = {
			name = "Request Sound",
			type = "group",
			guiInline = true,
			args = {
				help = {
					order = 0,
					type = "description",
					name = "Select a sound to play when one of your abilities is requested by another player",
				},
				Sound = {
                    order = 1,
                    type = "select",
                    width = 'normal',
                    name = "Sound",
                    desc = "Sound file to play when one of your abilities is requested by another player",
                    values = function()
							local t = {}							
							for k,v in pairs(sounds) do t[k] = k; end
							sort(t);
							return t;
					end,
					get = "GetRequestSound",
					set = "SetRequestSound"
                },
			}
		},
    }
}
local defaults = {
			char = {
				ButtonSizeLocal = 28,
				ButtonSizeForeign = 28,
				ButtonPadding = 4,
				MainFrameWidth = 200,
				MainFrameHeight = 400,
				MainFrameX = 0,
				MainFrameY = 0,
				NameTrimLength = 4,
				OpacityCD = 0.8,
				OpacityAvailable = 1,
				RequestSound = "Ding",
				Published = {} -- published abilities
			}
}

-- profile db pointer
local db

-------------------------------------
-------------------------------------
-- Settings handlers
function Telepathy:GetButtonSizeLocal(info)
	return db.ButtonSizeLocal;
end
function Telepathy:SetButtonSizeLocal(info, newval)
	db.ButtonSizeLocal = newval	
	for k,v in pairs(PublishedAbilities) do
		if v.button then
			v.button:SetHeight(newval)
			v.button:SetWidth(newval)
		end
	end	
	if self.MainPanel then
		self.MainPanel:Refresh()
	end
end
--
function Telepathy:GetButtonSizeForeign(info)
	return db.ButtonSizeForeign;
end
function Telepathy:SetButtonSizeForeign(info, newval)
	db.ButtonSizeForeign = newval
	for k,v in pairs(LinkedMembers) do
		for k1,info in pairs(v.Abilities) do
			if not info.Hidden and info.container and info.button then
				info.button:SetHeight(newval)
				info.button:SetWidth(newval)
				info.container:SetHeight(newval+14)
				info.container:SetWidth(newval)
			end
		end
	end
	if self.ForeignPanel then
		self.ForeignPanel:Refresh()
	end
end
--
function Telepathy:GetButtonPadding(info)
	return db.ButtonPadding;
end
function Telepathy:SetButtonPadding(info, newval)
	db.ButtonPadding = newval	
	if self.MainPanel then
		self.MainPanel:SetPadding(newval)
		self.MainPanel:Refresh()
	end
	if self.ForeignPanel then
		self.ForeignPanel:SetPadding(newval)
		self.ForeignPanel:Refresh()
	end
end
--
function Telepathy:GetNameTrimLength(info)
	return db.NameTrimLength
end
function Telepathy:SetNameTrimLength(info, newval)
	db.NameTrimLength = newval
end
--
function Telepathy:GetOpacityCD(info)
	return db.OpacityCD
end
function Telepathy:SetOpacityCD(info, newval)
	db.OpacityCD = newval
	opacityCD = newval

	for k,v in pairs(LinkedMembers) do
		for k1,info in pairs(v.Abilities) do
			if not info.Hidden and info.container and info.button and info.cd then
				if info.cd:IsShown() then
					info.container:SetAlpha(opacityCD)
					--info.cd:SetAlpha(opacityCD)
					--info.button:SetAlpha(opacityCD)
				end
			end
		end
	end
end
--
function Telepathy:GetOpacityAvailable(info)
	return db.OpacityAvailable
end
function Telepathy:SetOpacityAvailable(info, newval)
	db.OpacityAvailable = newval
	opacityAvailable = newval

	for k,v in pairs(LinkedMembers) do
		for k1,info in pairs(v.Abilities) do
			if not info.Hidden and info.container and info.button and info.cd then
				if not info.cd:IsShown() then
					info.container:SetAlpha(opacityAvailable)
					--info.cd:SetAlpha(opacityAvailable)
					--info.button:SetAlpha(opacityAvailable)
				end
			end
		end
	end
end
--
function Telepathy:GetRequestSound(info)
	return db.RequestSound
end
function Telepathy:SetRequestSound(info, newval)
	db.RequestSound = newval
	self:Trace("Setting request sound to", newval)

	self:Trace(info.arg)
	local sound_file = sounds[newval]	
	if sound_file and strlen(sound_file) > 0 then
		self:Trace("Playing soundfile", sound_file)
		PlaySoundFile(sound_file)		
	end
end
--/Settings Handlers
-------------------------------------
-------------------------------------


function Telepathy:Trace(...)
	if teleDebug then
		self:Print(...)
	end
end

function Telepathy:OnInitialize()
    -- Called when the addon is loaded
    -- Print a message to the chat frame
    self:Trace("OnInitialize Event Fired: Hello")
end

local function makeSizable(frame, frameleveloffset)
	local grip = CreateFrame("Frame", nil, frame)
	grip:EnableMouse(true)
	grip:SetFrameLevel(frame:GetFrameLevel() + (frameleveloffset or 0))
	local tex = grip:CreateTexture()
	grip.tex = tex
	tex:SetTexture([[Interface\BUTTONS\UI-AutoCastableOverlay]])
	tex:SetTexCoord(0.619, 0.760, 0.612, 0.762)
	tex:SetDesaturated(true)
	tex:ClearAllPoints()
	tex:SetPoint("TOPLEFT")
	tex:SetPoint("BOTTOMRIGHT", grip, "TOPLEFT", 19, -18)
	grip:SetWidth(22)
	grip:SetHeight(21)
	grip:SetScript("OnMouseDown", function(self)
		self:GetParent():StartSizing()
	end)
	grip:SetScript("OnMouseUp", function(self)
		self:GetParent():StopMovingOrSizing()
	end)

	grip:ClearAllPoints()
	grip:SetPoint("BOTTOMRIGHT")
	grip:SetScript("OnEnter", function(self)
		self.tex:SetDesaturated(false)
	end)
	grip:SetScript("OnLeave", function(self)
		self.tex:SetDesaturated(true)
	end)
end

local function sortContainerByClass(a, b)
	local r = true
	if a and b then
		local aMT, bMT, aI, bI = a.MemberTable, b.MemberTable, a.info, b.info
		if aMT and bMT then
			if aMT == bMT then
				r = aI.id < bI.id
			else
				if aMT.ClassName ~= bMT.ClassName then
					r = aMT.ClassName < bMT.ClassName
				else
					r = aMT.Name < bMT.Name
				end
			end
		elseif not aMT then
			r = true
		elseif not bMT then
			r = false
		end
	elseif not a then
		r = true		
	elseif not b then
		r = false
	end
	return r
end

local function SortButtonsByClassPlayer(t)
	sort(t, sortContainerByClass)

	--apply wrap hints based on class/ability
	local oldclass = nil
	for i, v in ipairs(t) do
		v.WrapHint = nil
		if v and v.MemberTable then
			if oldclass ~= v.MemberTable.ClassName then
				v.WrapHint = "NEWLINE"
				oldclass = v.MemberTable.ClassName
			end
		end
	end
end

local function round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end

function Telepathy.OnMainFrameSizeChanged(frame, w, h)
	--store in profile db
	db.MainFrameWidth = w
	db.MainFrameHeight = h

	local mp = Telepathy.MainPanel
	local fp = Telepathy.ForeignPanel
	local width = w
	local heightmp = mp:GetHeight() -- this one auto sizes, so take it's height into account
	local heightfp = h - heightmp
	if mp and mp:GetWidth() ~= width then
		mp:SetWidth(width)		
		mp:Refresh()
	end
	if fp and fp:GetWidth() ~= width then
		--Telepathy:Trace("Setting fp width to",width)
		fp:SetWidth(width)
		fp:Refresh()
	end
	if fp and fp:GetHeight() ~= heightfp then
		fp:SetHeight(heightfp)
		fp:Refresh()
	end
end

function Telepathy:StoreMainFramePosition()
	db.MainFrameX = self.MainFrame:GetLeft()
	db.MainFrameY = self.MainFrame:GetTop()
end

function Telepathy:CreateMainFrame()

	local f = CreateFrame("Frame", "TelepathyMainFrame", UIParent)
	f:ClearAllPoints()
	if db.MainFrameX == 0 and db.MainFrameY == 0 then
		f:SetPoint("CENTER",UIParent)
	else
		f:SetPoint("TOPLEFT",UIParent,"BOTTOMLEFT",db.MainFrameX,db.MainFrameY)
	end
	local buttonSize = db.ButtonSizeLocal
	f:EnableMouse(true)	
	f:SetMovable(true)
	f:SetResizable(true)
	f:SetMinResize(3*buttonSize,4*buttonSize)
	f:SetWidth(db.MainFrameWidth)
	f:SetHeight(db.MainFrameHeight)
	f:RegisterForDrag("LeftButton")
	f:SetScript("OnDragStart",f.StartMoving)
	f:SetScript("OnDragStop",function() f:StopMovingOrSizing(); Telepathy:StoreMainFramePosition(); end)
	f:SetScript("OnSizeChanged", Telepathy.OnMainFrameSizeChanged)
	f:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 8, 
		insets = { left = 2, right = 2, top = 2, bottom = 2}
	})	
	f:SetBackdropBorderColor(0.2,0.2,0.2)
	f:SetBackdropColor(24/255, 24/255, 24/255, 0.2)
	f:Show()

	makeSizable(f, 2)

	self.MainFrame = f
end

function Telepathy:CreateMainPanel()
	local parent = self.MainFrame
	local f = CreateWrapPanel(parent)
	f:SetScript("OnMouseDown", function(self, button)						
			  if button == "LeftButton" and not parent.isMoving then
			   parent:StartMoving();
			   parent.isMoving = true;
			  end
			end)
	f:SetScript("OnMouseUp", function(self, button)
			  if button == "LeftButton" and parent.isMoving then
			   parent:StopMovingOrSizing();
			   Telepathy:StoreMainFramePosition();
			   parent.isMoving = false;
			  end
	end)
	local buttonSize = db.ButtonSizeLocal
	f.Tag = "MainPanel"
	f:SetPadding(db.ButtonPadding)
	f:SetMargins(5)
	f:ClearAllPoints()
	f:SetPoint("BOTTOMLEFT",self.MainFrame)
	f:EnableMouse(true)		
	f:SetHeight(buttonSize+buttonPadding)	
	f.MinHeight = buttonSize+buttonPadding
	f.AutoSizeY = true	
	f:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 8, 
		insets = { left = 2, right = 2, top = 2, bottom = 2}
	})	
	f:SetBackdropBorderColor(0.2,0.2,0.2)
	f:SetBackdropColor(24/255, 24/255, 24/255, 0.2)
	f:Show()
	self.MainPanel = f
	f:SetContext(ButtonCache.Local)
end

function Telepathy:CreateForeignPanel()
	local parent = self.MainFrame
	local f = CreateWrapPanel(parent)
	f:SetScript("OnMouseDown", function(self, button)						
			  if button == "LeftButton" and not parent.isMoving then
			   parent:StartMoving();
			   parent.isMoving = true;
			  end
			end)
	f:SetScript("OnMouseUp", function(self, button)
			  if button == "LeftButton" and parent.isMoving then
			   parent:StopMovingOrSizing();
			   Telepathy:StoreMainFramePosition();
			   parent.isMoving = false;
			  end
			end)
	f.Tag = "ForeignPanel"
	f:SetPadding(db.ButtonPadding)
	f:SetMargins(5)
	f:ClearAllPoints()
	f:SetPoint("TOPLEFT",parent,"TOPLEFT",0,0)
	f:EnableMouse(true)
	--[[
	f:SetScript("OnUpdate", function(frame, elapsed)
		frame.UpdateDelay = elapsed + (frame.UpdateDelay or 0)
		if frame.UpdateDelay > 0.5 then
			local updateTable = frame.MemberTablesForUpdate
			if updateTable and #updateTable > 0 then
				local t = updateTable[#updateTable]
				Telepathy:RefreshForeignPanel(t)
				updateTable[#updateTable] = nil
				if #updateTable == 0 then
					frame.UpdateDelay = 0 --no more processing
				end
			end			
		end
	end)
	f.MarkTableForUpdate = function(frame, t)
		local MemberTablesForUpdate = frame.MemberTablesForUpdate or {}
		frame.MemberTablesForUpdate = MemberTablesForUpdate
		local bNeedsUpdate = true
		for k,v in pairs(MemberTablesForUpdate) do
			if v == t then
				bNeedsUpdate = false
				break
			end
		end
		if bNeedsUpdate then
			tinsert(MemberTablesForUpdate, t)
		end
	end
	]]
	local buttonSize = db.ButtonSizeForeign
	local buttonPadding = db.ButtonPadding
	f:SetHeight((buttonSize+buttonPadding)*4+borderPadding)
	f.MinHeight = buttonSize+buttonPadding+borderPadding	
	f:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 8, 
		insets = { left = 2, right = 2, top = 2, bottom = 2}
	})	
	f:SetBackdropBorderColor(0.2,0.2,0.2)
	f:SetBackdropColor(24/255, 24/255, 24/255, 0.2)
	f:Show()
	f:SetContext(ButtonCache.Foreign)
	self.ForeignPanel = f
end

local function AttachTooltips(frame, info)
	if info.type == "spell" then
		frame:HookScript("OnEnter", function(self) GameTooltip:SetOwner(self); GameTooltip:SetSpellByID(info.id); end )
	elseif itemid then
		frame:HookScript("OnEnter", function(self) GameTooltip:SetOwner(self); GameTooltip:SetItemByID(info.id); end )
	end
	frame:HookScript("OnLeave", function() GameTooltip:Hide(); end )
end

function Telepathy:CreateSpellBook()
	
	if Telepathy_SpellBook then
		if Telepathy_SpellBook:IsVisible() then
			Telepathy_SpellBook:Hide()
			Telepathy_SpellBook = nil --destroy
			return
		end		
	end	

	local buttonSize = db.ButtonSizeForeign
	local wrap
	local f = CreateFrame("Frame", "TelepathySpellBook", UIParent)
	Telepathy_SpellBook = f
	f:SetFrameLevel(Telepathy.MainFrame:GetFrameLevel() + 5)
	f:ClearAllPoints()
	f:SetPoint("CENTER",UIParent)
	f:EnableMouse(true)	
	f:SetMovable(true)
	f:SetResizable(true)
	f:SetMinResize(6*buttonSize,4*buttonSize)
	f:SetWidth(500)
	f:SetHeight(400)
	f:RegisterForDrag("LeftButton")
	f:SetScript("OnDragStart",f.StartMoving)
	f:SetScript("OnDragStop",f.StopMovingOrSizing)	
	f:SetBackdrop({
		bgFile = [[Interface\AchievementFrame\UI-Achievement-Parchment-Horizontal]], tile = false, tileSize = 16,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 16, 
		insets = { left = 4, right = 4, top = 4, bottom = 4}
	})	
	f:SetBackdropBorderColor(0.2,0.2,0.2)
	f:Show()
	makeSizable(f, 5)

	--close button
	local closebutton = CreateFrame("Button", nil, f, "UIPanelButtonTemplate")
	closebutton:ClearAllPoints()	
	closebutton:SetHeight(24)
	closebutton:SetWidth(70)
	closebutton:SetText("Close")	
	closebutton:Show()
	closebutton:SetScript("OnClick", function(b) Telepathy:CreateSpellBook() end)
	--close button position set relative to the lower wrap panel below

	--hint label - center of spellbook
	local hint1 = f:CreateFontString(nil,"ARTWORK")
	hint1:SetFont("Fonts\\ARIALN.TTF",11) --(GameFontNormal:GetFont())
	hint1:SetText([[|cffCED000Left-click a player's abilitiy to add it to the Combat Window. Right-click to remove.|r]])
	hint1:SetPoint("CENTER")
	hint1:SetWidth(120)
	hint1:Show()

	--foreign wrap panel
	wrap = CreateWrapPanel(f)
	wrap:ClearAllPoints()
	wrap:SetPoint("TOPLEFT")
	wrap:SetPoint("TOPRIGHT")
	wrap:EnableMouse(true)
	wrap:SetScript("OnMouseDown", function(self, button)						
			  if button == "LeftButton" and not f.isMoving then
			   f:StartMoving();
			   f.isMoving = true;
			  end
			end)
	wrap:SetScript("OnMouseUp", function(self, button)
			  if button == "LeftButton" and f.isMoving then
			   f:StopMovingOrSizing();
			   f.isMoving = false;
			  end
	end)
	wrap.AutoSizeY = true
	wrap:SetMargins(10)
	wrap:SetPadding(5)
	wrap:Show()

	local context = {}
	local button, container

	local f_OnForeignClick = function(b, mousebutton)
					local info = b.info
					Telepathy:Trace("OnClick on spellbook spell:", mousebutton)
					if mousebutton == "LeftButton" then
						info.Hidden = false
						PlaySoundFile("Sound\\interface\\uSpellIconPickup.wav")
					elseif mousebutton == "RightButton" then
						info.Hidden = true
						PlaySoundFile("Sound\\interface\\uSpellIconDrop.wav")
					end
					Telepathy:RefreshForeignPanel(info.MemberTable)
					--Telepathy.ForeignPanel:MarkTableForUpdate(info.MemberTable)
				end

	f.RepopulateForeignButtons = function()
		local context = wrap.Context or {}
		for k,v in pairs(context) do v:Hide(); v = nil; end
		wipe(context)
		
		for name, membertable in pairs(LinkedMembers) do
			for k,v in pairs(membertable.Abilities) do
				button, _, container = self:CreateForeignAbilityButton(v, wrap, false, membertable.Name)
				self:Trace("Adding container at index",#context + 1,container)
				tinsert(context,container)
				container.MemberTable = membertable
				container.info = v
				button.info = v			
				button:SetScript("OnClick", f_OnForeignClick)
			end
		end
		SortButtonsByClassPlayer(context)
		wrap:SetContext(context)
		wrap:Refresh()
	end


	f.RepopulateForeignButtons()
	
	--create the lower panel for my own spells
	local mycontext = {}
	local p = CreateWrapPanel(f)
	p:ClearAllPoints()
	p:SetPoint("BOTTOMLEFT")
	p:SetPoint("BOTTOMRIGHT")
	p:EnableMouse(true)
	p:SetScript("OnMouseDown", function(self, button)						
			  if button == "LeftButton" and not f.isMoving then
				 local infoType, info1, info2 = GetCursorInfo()
				 if not infoType then
					f:StartMoving();
					f.isMoving = true;
				 end
			  end
			end)
	p:SetScript("OnMouseUp", function(self, button)
				if button == "LeftButton" then
					local infoType, info1, info2 = GetCursorInfo()
					if not infoType then
						if f.isMoving then
							f:StopMovingOrSizing();
							f.isMoving = false;
						end
					else
						if (infoType and infoType == "spell") then												
							local spell, id = GetSpellBookItemInfo(info1, "spell")
							Telepathy:Trace("Drop on spellbook:", infoType, info1, spell)						
							Telepathy:PublishAbility(id)
							ClearCursor()
							p.RepopulateButtons()
						else
							Telepathy:Trace("Mouse up on spellbook, cursor info type=", infoType)
						end						
					end
			  end
	end)	
	p.AutoSizeY = true
	p:SetMargins(10)
	p:SetPadding(5)
	p:SetHeight(buttonSize + 10)
	p.MinHeight = p:GetHeight()
	p:SetWidth(f:GetWidth())
	p:SetFrameLevel(f:GetFrameLevel()+5)
	p:SetBackdrop({
		--bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 16,
		bgFile = [[Interface\AchievementFrame\UI-Achievement-Parchment-Horizontal]], tile = false, tileSize = 16,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 16, 
		insets = { left = 4, right = 4, top = 4, bottom = 4}
	})	
	p:SetBackdropBorderColor(0.2,0.2,0.2)
	p:Show()

	--close button relative to this panel's top right corner
	closebutton:SetPoint("BOTTOMRIGHT", p, "TOPRIGHT", -5, 0)

	--populate context
	p.RepopulateButtons = function()

		local f_OnMyClick = function(b, mousebutton)
			if mousebutton == "RightButton" and b.info then
				local id = b.info.id
				Telepathy:PublishAbility(id, true)
				PublishedAbilities[id] = nil
				PlaySoundFile("Sound\\interface\\uSpellIconDrop.wav")
				p.RepopulateButtons()
			end			
		end

		local context = p.Context or {}
		--clear context
		for k,v in pairs(context) do v:Hide() end
		wipe(context)

		for k,v in pairs(PublishedAbilities) do
			local info = v
			local start, duration, enabled
			local b = CreateFrame("Button", nil, p)
			local texture
			if info.type == "spell" then
				local _, _, spellicon = GetSpellInfo(info.id)	
				texture = spellicon
			elseif info.type == "item" then
				local _, _, _, _, _, _, _, _, _, itemicon = GetItemInfo(info.id)
				texture = itemicon
			end   	 	
			b:SetNormalTexture(texture)
			b:SetHeight(db.ButtonSizeLocal)
			b:SetWidth(db.ButtonSizeLocal)
			b:Show()
			b:SetScript("OnClick", f_OnMyClick)
			b:RegisterForClicks("RightButtonUp")
			b.info = info

			AttachTooltips(b, info)

			tinsert(context, b)
		end
		p:SetContext(context)
		p:Refresh()
	end

	p.RepopulateButtons()
	f.RepopulateButtons=p.RepopulateButtons

	--hint label - center of my wrap panel
	local hint2 = p:CreateFontString(nil,"ARTWORK")
	hint2:SetFont("Fonts\\ARIALN.TTF",11) --(GameFontNormal:GetFont())
	hint2:SetText([[|cffCED000Drag an ability onto this pane to publish it. Right-click to remove.|r]])
	hint2:SetPoint("CENTER")	
	hint2:Show()

	f:SetScript("OnSizeChanged", 
		function(self, w, h)
			if wrap:GetWidth() ~= w then
				wrap:SetWidth(w)				
			end
			if wrap:GetHeight() ~= (h - p:GetHeight()) then
				wrap:SetHeight(h - p:GetHeight())
			end
			if p:GetWidth() ~= w then
				p:SetWidth(w)
			end
			p:Refresh()
			wrap:Refresh()
		end
	)
	
end

local function ClassColor(name)
	local class, classname = nil, "UNKNOWN"
	if UnitExists(name) then		
		class, classname = UnitClass(name)
		if not classname then
			classname = "UNKNOWN" --should really be SHAMAN, cos shaman are filled with win
		end
	end
	if teleDebug then
		classname = LinkedMembers[name].ClassName
	end
	local c = Colors["Class"][classname]
	return c[1], c[2], c[3]
end

function Telepathy:CreateForeignPlayerPanel(parentframe, membertable, forspellbook)
	if not forspellbook and membertable.Panel then
		return membertable.Panel
	end

	--self:Trace("Creating player panel for", membertable.Name)

	local panel = CreateWrapPanel(parentframe)
	panel.Tag = "TelepathyForeign-"..membertable.Name
	panel:ClearAllPoints()
	panel:SetPoint("TOPLEFT",parentframe)	
	local context = {} -- button list
	if not forspellbook then
		membertable.Panel = panel
		membertable.PanelButtons = context
	end

	--self:Trace("Player panel created")

	panel.AutoSizeY = true
	panel.AutoSizeX = true
	panel.MinHeight = buttonSize	
	panel:SetPadding(4,2)
	panel:SetMargins(5)
	panel.BGTexture=panel:CreateTexture()
	panel.BGTexture:SetAllPoints(panel)
	panel.BGTexture:SetTexture(0.1,0.1,0.1,0.3)

	--create the name of the player here
	--self:Trace("Creating name label")
	local nameLabel = panel:CreateFontString(nil,"ARTWORK")
	nameLabel:SetFont(GameFontNormal:GetFont())
	--self:Trace("Name label created, setting text")
	nameLabel:SetText(membertable.Name)
	local r,g,b =ClassColor(membertable.Name)
	nameLabel:SetTextColor(r,g,b,1)
	nameLabel.WrapYOffset = 3
	panel.MinWidth = nameLabel:GetWidth() + buttonSize + buttonPadding
	self:Trace("Set player panel minwidth=",panel.MinWidth)
	--self:Trace("Inserting name label into context")
	tinsert(context,nameLabel)
	--now all of the published abilities (or just the shown ones, depending on frame type)
	local button, cooldown, container
	local createcooldown = (not forspellbook)
	for k,v in pairs(membertable.Abilities) do
		if forspellbook then
			--self:Trace("Adding foreign ability button")
			button, cooldown, container = self:CreateForeignAbilityButton(v, panel, createcooldown)			
			--self:Trace("Inserting into table")
			tinsert(context,container)
			--self:Trace("Insert done")
			if not forspellbook then
				v.button, v.cd = button, cooldown
			end
			--self:Trace("Moving on...")
		end
	end		
	--self:Trace("Showing panel")
	panel:Show()
	--self:Trace("Setting panel context")
	panel:SetContext(context, true)	
	local prefWidth = panel:PreferredWidth()
	--self:Trace("Getting preferred width returned", prefWidth)
	panel:SetWidth(prefWidth)
	--self:Trace("Refreshing panel")
	panel:Refresh()	

	if not forspellbook then
		tinsert(ButtonCache.Foreign, #ButtonCache.Foreign, panel)
		Telepathy.ForeignPanel:Refresh()
	end
	return panel
end

local function removeentry(tab, val)
	for k, v in pairs(tab) do
		if v == val then
			tab[k] = nil
		end
	end
end

local function RequestAction(button)
	local info = button.info
	if not info then return end
	if not info.requestable then return end
	if info.cd:IsVisible() then
		Telepathy:Trace("Not requesting action, as it is still on CD!")
		return
	end
	
	PlaySoundFile("Sound\\interface\\iUiInterfaceButtonA.wav")
	Telepathy:ComRequest("a", info.id, info.type, info.Name)
end

function Telepathy:RefreshForeignPanel(membertable)
	
	if not membertable then return end
	--if not membertable.Panel then return end
		
	self:Trace("RefreshForeignPanel(membertable=",membertable.Name,")")	

	--local context = membertable.PanelButtons
	local panel = self.ForeignPanel --membertable.Panel		
	local fcache = ButtonCache.Foreign
	local button, cooldown, container

	for k,v in pairs(membertable.Abilities) do
		if not v.Hidden and not v.button then			
			button, cooldown, container = self:CreateForeignAbilityButton(v, panel, true, membertable.Name)
			self:Trace("Adding container at index",#fcache + 1,container)
			tinsert(fcache,container)
			v.button, v.cd, v.container = button, cooldown, container
			container.MemberTable = membertable
			container.info = v
			self:RefreshSpellCooldown(v)
		elseif v.button and v.Hidden then
			self:Trace("Removing container from foreign cache")
			removeentry(fcache, v.container)
			v.container:Hide()
			v.button = nil --throw the button and cd away
			v.cd = nil
			v.container = nil
		end
	end
	--local prefwidth = panel:PreferredWidth()
	--self:Trace("Preferred width =",prefwidth)
	--panel:SetWidth(min(prefwidth, self.ForeignPanel:GetWidth()))
	SortButtonsByClassPlayer(fcache)
	panel:Refresh()
end

local function OnCDButtonUpdate(self)
	if self.needsupdate then
		self.needsupdate = false
		local i = self.info
		local cd = i.cd

		local start, duration, enabled
		if i.type == "spell" then
			start, duration, enabled = GetSpellCooldown(i.id)
		elseif i.type == "item" then
			start, duration, enabled = GetItemCooldown(i.id)
		end
		cd:SetCooldown(start, duration)

		Telepathy:ComNotifyCD(i.id, i.type, currentChannel, nil)
		if teleDebug then
			Telepathy:ComNotifyCD(i.id, i.type, "WHISPER", UnitName("player"))
		end
	end
end

function Telepathy:RegisterForeignAbility(id, type, requestable, from, cd_remaining, cd_duration, suppress_refresh)
	local member = LinkedMembers[from] or { Name=from, Abilities={}}
	LinkedMembers[from] = member
	local info = member.Abilities[id] or { id=id, type=type, requestable=requestable, Name=from, Hidden=true }
	member.Abilities[id] = info
	member.Name = from	
	info.MemberTable = member
	info.cd_duration = cd_duration
	info.cd_start = GetTime() + cd_remaining - cd_duration

	if UnitExists(from) then
		_, member.ClassName = UnitClass(from)
	elseif teleDebug then
		--random class
		local i, j = random(1, 10), 1
		for k,v in pairs(Colors["Class"]) do
			if i == j then
				member.ClassName = k
				break
			end
			j = j + 1
		end
	end
	
	if Telepathy_SpellBook and not suppress_refresh then
		Telepathy_SpellBook.RepopulateForeignButtons()
	end
end

function Telepathy:CreateForeignAbilityButton(info, parentframe, includecooldown, playername)
	local id = info.id
	local t = info.type
	local start, duration, enabled	
	local buttonSize = db.ButtonSizeForeign
	
	--create a bottom frame for the player name
	local f = CreateFrame("Frame", nil, parentframe)
	f:SetHeight(buttonSize + 14) --name at the bottom
	f:SetWidth(buttonSize)
	f:Show()
	
	local trimlen = db.NameTrimLength
	if trimlen == 0 then trimlen = 99 end
	local nameforbutton = playername
	if playername and #playername > trimlen then
		nameforbutton = strsub(playername, 1, trimlen)
	end
	
	local nameLabel = f:CreateFontString(nil,"ARTWORK")
	nameLabel:SetFont("Fonts\\ARIALN.TTF",11, "OUTLINE") --(GameFontNormal:GetFont())
	nameLabel:SetText(nameforbutton)
	self:Trace("Set player label text to",nameforbutton)
	local r,g,b =ClassColor(playername)
	nameLabel:SetTextColor(r,g,b,1)
	nameLabel:SetPoint("BOTTOMLEFT", f)
	nameLabel:SetPoint("BOTTOMRIGHT", f)
	nameLabel:Show()
	
	local b = CreateFrame("Button", nil, f)
	b:SetPoint("TOP")
	b.id = id
	b.info = info	
	local texture
	if t == "spell" then
		local _, _, spellicon = GetSpellInfo(id)	
		texture = spellicon
	elseif t == "item" then
		local _, _, _, _, _, _, _, _, _, itemicon = GetItemInfo(id)
		texture = itemicon
	end   	 	
	b:SetNormalTexture(texture)
	b:SetHeight(buttonSize)
	b:SetWidth(buttonSize)
	b:Show()
	b:RegisterForClicks("AnyUp")
	AttachTooltips(b, info)

	local cd = nil
	if includecooldown then
		--only adding the RequestAction script to buttons which have a cooldown - the combat window buttons
		b:SetScript("OnClick", RequestAction)
		--

		cd = CreateFrame("Cooldown",nil,b)
		cd:SetAllPoints()
		cd:SetScript("OnHide", function(self)
			b:SetAlpha(opacityAvailable);
			self:SetAlpha(opacityAvailable);
		end)
		cd:SetScript("OnShow", function(self)
			b:SetAlpha(opacityCD);
			self:SetAlpha(opacityCD);
		end)
	end
	return b, cd, f
end

function Telepathy:CreateMyAbilityButton(info)
	
	if PublishedAbilities[info.id] and PublishedAbilities[info.id].button then
		return PublishedAbilities[info.id].button
	end
	
	local buttonSize = db.ButtonSizeLocal
	local start, duration, enabled
	local mp = self.MainPanel
	local b = CreateFrame("Button", nil, mp, "SecureActionButtonTemplate")
	b:SetScript("OnUpdate", OnCDButtonUpdate)
	b:SetAttribute("type", info.type)	
	b:SetAttribute("id", info.id)
	--b:SetFrameLevel(mp:GetFrameLevel()+10)
	b.id = info.id
	b.info = info
	if (info.type == "spell") then				
		b:SetAttribute("spell", info.name)
		start, duration, enabled = GetSpellCooldown(info.id)
	end
	if (info.type == "item") then
		b:SetAttribute("item", info.name)
		start, duration, enabled = GetItemCooldown(info.id)
	end
	b:ClearAllPoints()
	b:SetPoint("TOPLEFT",mp,"TOPLEFT",0,0)
	local texture
	if info.type == "spell" then
		local _, _, spellicon = GetSpellInfo(info.id)	
		texture = spellicon
	elseif info.type == "item" then
		local _, _, _, _, _, _, _, _, _, itemicon = GetItemInfo(info.id)
		texture = itemicon
	end   	 	
	b:SetNormalTexture(texture)
	b:SetHeight(buttonSize)
	b:SetWidth(buttonSize)
	b:Show()

	local cd = CreateFrame("Cooldown",nil,b)
	cd:SetAllPoints()
	cd:SetCooldown(start, duration)			
	info.button = b
	info.cd = cd	
	
	AttachTooltips(b, info)

	local highlight = b:CreateTexture()
	highlight:SetHeight(b:GetHeight() + (buttonPadding*2))
	highlight:SetWidth(b:GetWidth() + (buttonPadding*2))
	highlight:ClearAllPoints(b)
	highlight:SetPoint("CENTER")
	highlight:SetTexture(0,0,0,0)
	info.highlight = highlight

	--add to main panel?
	tinsert(ButtonCache.Local, b)
	mp:Refresh()
end

function GetCooldownRemaining(id)
	local started, duration, a = GetSpellCooldown(id)
	return (started + duration - GetTime()), started, duration
end

function Telepathy:PublishAbility(abilityname, bRemove, bDeferUpdate)
	self:Trace("PublishAbility("..abilityname..")")
	
	local Id
	if type(abilityname) == "string" then
		local _, _, Color, Ltype, Id, Enchant, Gem1, Gem2, Gem3, Gem4, Suffix, Unique, LinkLvl, Name = string.find(abilityname, "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")	
		if Id then
			self:Trace(Id, ":", Name)
			abilityname = Id	
		end
	else
		Id = abilityname
	end

	local atype
	local info = GetSpellInfo(Id)
	if info then atype = "spell" else atype = "item" end
	if not info then info = GetItemInfo(Id) end
	if not info then return end -- dunno what this is
	local id = tonumber(Id)

	--what are we publishing?
	if not bRemove then		
		local p = PublishedAbilities[id] or {}
		PublishedAbilities[id] = p
	
		p.type = atype
		p.name = info
		p.id = id
		p.requestable = true

		self:Trace(p.name, p.type, p.id)

		local useable, mana = IsUsableSpell(abilityname)
		self:Trace("useable :", useable, " lowmana :", mana)

		self:CreateMyAbilityButton(p)
	else
		local info = PublishedAbilities[Id]
		if info.button then
			info.button:Hide()
		end
		removeentry(ButtonCache.Local, info.button)
		PublishedAbilities[Id] = nil
		self.MainPanel:Refresh()
	end

	if not bDeferUpdate then
		local cd_remaining, _, duration = GetCooldownRemaining(id)
		local pubstring = strconcat(id,"|",atype:sub(1,1),"|","y","|",round(cd_remaining,2),"|",duration)
		if bRemove then
			pubstring = pubstring.."|-"
		end	
		self:ComPublish(pubstring, currentChannel, nil)
	end
end

function Telepathy:UNIT_SPELLCAST_SUCCEEDED(event, ...)
	local unit, _, _, _, spellid = ...
	if UnitIsPlayer(unit) then
		local pubinfo = PublishedAbilities[spellid]
		if pubinfo and pubinfo.button then
			pubinfo.button.needsupdate = true
			if pubinfo.highlight then
				pubinfo.highlight:Hide()				
			end
			self:Trace("set pubinfo.button.needsupdate = true")
		end
	end
end

function Telepathy:PARTY_MEMBERS_CHANGED(event,...)
	if not self.PartyChanging then
		self.PartyChanging = true
		self:Trace("PARTY_MEMBERS_CHANGED: starting update timer")
		self:ScheduleTimer("OnPartyChangedTimer", 1, nil)
	end
end

local function GetCurrentChannel()
	local i = GetNumRaidMembers()
	if i > 0 then
		currentChannel = "RAID"
	else
		i = GetNumPartyMembers()
		if i > 0 then
			currentChannel = "PARTY"
		end		
	end
	return currentChannel, i
end

function Telepathy:OnPartyChangedTimer()
	self.PartyChanging = false	
	local c, playercount = GetCurrentChannel()
	c = c:lower() --lower channel
	if c == "battleground" then c = "raid" end
	
	self:Trace("OnPartyChangedTimer: new channel=", c, "with", playercount, "players")

	--loop players
	local existing = {}
	local name
	for i=1,playercount do
		name = UnitName(c..i)
		if name then
			existing[name] = true
			if LinkedMembers[name] then
				--I have met you already
			else
				--I don't know you, please share your published abilities?
				self:Trace("Sending refresh request to new person:", name)
				self:ComRequest("r", 0, 0, name) 
			end
		end
	end
	--remove anyone who we don't have anymore
	for k,v in pairs(LinkedMembers) do
		if not existing[k] then
			local a = v.Abilities
			for k1, info in pairs(a) do
				info.Hidden = true				
			end
			self:RefreshForeignPanel(v)
			LinkedMembers[k] = nil	
		end
	end
		
	if Telepathy_SpellBook then
		Telepathy_SpellBook.RepopulateForeignButtons()
	end
end


function Telepathy:HandleSlashCommand(args, info)
	if args == "?" then
		self:Print("Usage: /telepathy -- shows the spell book")
		--self:Print("Usage: /tele  -- shows the spell book")
		--self:Print("Usage: /tele pub [spelllink/itemlink] -- publish a spell/item")				
	else
		args = args:trim()
		if args:find("pub") then
			local a = args:sub(4):trim()			
			self:PublishAbility(a)
		elseif args:find("req") then
			local r = args:sub(4):trim()
			self:OnComRequest(r)		
		elseif args == "reset" then
			--reset the position of the windows
			db.MainFrameX = defaults.char.MainFrameX
			db.MainFrameY = defaults.char.MainFrameY
			db.MainFrameWidth = defaults.char.MainFrameWidth
			db.MainFrameHeight = defaults.char.MainFrameHeight						
			Telepathy:CreateCombatFrames()
			Telepathy:Print("Combat window has been reset - please reload your UI to complete the reset.")
		elseif args == "debug" then
			--toggle debugging
			teleDebug = not teleDebug
			if teleDebug then
				self:Print("Debugging is ON")
			else
				self:Print("Debugging is OFF")
			end
		else		
			Telepathy:CreateSpellBook()
		end
	end
end

function Telepathy:HandleSlashCommandTest(args, info)
	--[[
	args = args:trim()
	Telepathy:Print(args)
	local s = tostring(args)
	local id, type, r, x = strsplit("|", s)
	Telepathy:Print(id, type, r, x)
	Telepathy:SendCommMessage(comPublish, s, "WHISPER", UnitName("player"))
	]]
	Telepathy:ComRefresh("WHISPER", UnitName("player"))
end

function Telepathy:HandleSlashCommandLoadTest(args, info)
	local sender = "Player"..random(1, 10)
	for i=20, random(22,25) do
		local spell, id = GetSpellBookItemInfo(i, "spell")
		if spell then
			local fakemessage = id.."|s|y|0|120"
			Telepathy:OnComPublish(fakemessage, "WHISPER", sender)
		end
	end

	--Telepathy.ForeignPanel:MarkTableForUpdate(LinkedMembers[sender])
	Telepathy:RefreshForeignPanel(LinkedMembers[sender])

	sender = UnitName("player")
	for i=20, random(22,27) do
		local spell, id = GetSpellBookItemInfo(i, "spell")
		if spell then
			local fakemessage = id.."|s|y|0|120"
			Telepathy:OnComPublish(fakemessage, "WHISPER", sender)
		end
	end

	--Telepathy.ForeignPanel:MarkTableForUpdate(LinkedMembers[sender])
	Telepathy:RefreshForeignPanel(LinkedMembers[sender])
end

function Telepathy:CreateSlashCommands()
	self:RegisterChatCommand("tele", "HandleSlashCommand")
	self:RegisterChatCommand("telepathy", "HandleSlashCommand")
	self:RegisterChatCommand("teletest", "HandleSlashCommandTest")
	self:RegisterChatCommand("teleloadtest", "HandleSlashCommandLoadTest")	
end

function Telepathy:RefreshSpellCooldown(info)
	if info.button and info.cd then		
		self:Trace("Setting spell cooldown:", info.cd_start, info.cd_duration)
		info.cd:SetCooldown(info.cd_start, info.cd_duration)
	end
end

function Telepathy:ComNotifyCD(id, type, channel, to)
	local start, duration, enabled
	if type == "spell" then
		start, duration, enabled = GetSpellCooldown(id)
	elseif type == "item" then
		start, duration, enabled = GetItemCooldown(id)
	end	
	local realstart = round(GetTime() - start, 3) --the number of seconds ago it started
	local commsg = strconcat(id,"|",type,"|",realstart,"|",duration)
	local info = PublishedAbilities[id]
	info.cd_duration = duration
	info.cd_start = start
	self:SendCommMessage(comCD, commsg, channel, to)
end

function Telepathy:OnComNotifyCD(msg, channel, sender)
	if sender == UnitName("player") then
		if not teleDebug then return end
	end

	Telepathy:Trace("Got Telepathy:OnComNotifyCD message: ", msg,", dist=",channel,", sender=", sender)
	
	local member = LinkedMembers[sender]	
	if not member then
		if teleDebug then
			Telepathy:Trace(sender,"has no table in LinkedMembers")
		end
		return
	end
	local tid, type, start, duration = strsplit("|", msg)
	local id = tonumber(tid)
	
	local info = member.Abilities[id]
	if not info then
		if teleDebug then
			Telepathy:Trace(sender,"has no info for id", id)
		end		
		return
	end

	Telepathy:Trace("Got foreign info for "..sender)
	local t = GetTime() - tonumber(start)
	info.cd_start = t
	info.cd_duration = tonumber(duration)
	Telepathy:RefreshSpellCooldown(info)
end

function Telepathy:ComPublish(msg, channel)
	Telepathy:SendCommMessage(comPublish, msg, channel, UnitName("player"))
	--Telepathy:SendCommMessage(comPublish, msg, "WHISPER", UnitName("player"))
end

function Telepathy:OnComPublish(msg, channel, sender)
	if sender == UnitName("player") then
		if not teleDebug then return end
	end

	Telepathy:Trace("Got Telepathy:OnComPublish message: ", msg,", dist=",channel,", sender=", sender)
	-- in the format: [spellid/itemid]|[s/i]|[y/n]
	local id, type, r, cd_remain, duration, delete = strsplit("|", msg)
	if type == "s" then
		type = "spell"
	elseif type == "i" then
		type = "item"
	end
	local requestable = (r == "y")

	if delete == "-" then
		--removing this abilitiy
		local mt = LinkedMembers[sender]
		local id = tonumber(id)
		if mt and mt.Abilities then
			mt.Abilities[id].Hidden = true
			--Telepathy.ForeignPanel:MarkTableForUpdate(mt)
			Telepathy:RefreshForeignPanel(mt)
			mt.Abilities[id] = nil
			if Telepathy_SpellBook then
				Telepathy_SpellBook.RepopulateForeignButtons()
			end
		end
	else
		Telepathy:RegisterForeignAbility(tonumber(id), type, requestable, sender, tonumber(cd_remain), tonumber(duration))
	end
end

-- send out a complete table of my abilities
function Telepathy:ComRefresh(channel, to)
	local sendtable = {}
	for k,v in pairs(PublishedAbilities) do
		local remain, _, duration = GetCooldownRemaining(v.id)
		local s = { id = v.id, r = v.requestable, t = v.type, cd = remain, d = duration }
		tinsert(sendtable, s)
	end
	local msg = self:Serialize(sendtable)
	Telepathy:Trace("Sending ComRefresh:", msg)
	Telepathy:SendCommMessage(comRefresh, msg, channel, to)
end

function Telepathy:OnComRefresh(msg, channel, sender)
	if sender == UnitName("player") then
		if not teleDebug then return end
	end

	Telepathy:Trace("Received ComRefresh from", sender, ":", msg)

	-- the  refresh message is a table with all published ability info in the form
	-- { id, type, requestable }
	local ok, data = Telepathy:Deserialize(msg)
	if ok and type(data) == "table" then
		local addedabilities = {}
		for k,v in pairs(data) do			
			Telepathy:Trace("ComRefresh: Registering foreing ability for", sender)
			Telepathy:RegisterForeignAbility(v.id, v.t, v.r, sender, v.cd, v.d, true)
			addedabilities[v.id] = true
		end

		if LinkedMembers[sender] then
			local mt_abilities = LinkedMembers[sender].Abilities

			-- remove any abilties that aren't there anymore
			for k,v in pairs(mt_abilities) do
				if not addedabilities[k] then
					Telepathy:Trace("Removing foreing ability from", sender, k)
					if v.container then 
						removeentry(ButtonCache.Foreign, v.container)
						v.container:Hide() 
					end
					mt_abilities[k] = nil
				end
			end
		end

		if Telepathy_SpellBook then
			Telepathy_SpellBook.RepopulateForeignButtons()
		end
		--Telepathy.ForeignPanel:MarkTableForUpdate(LinkedMembers[sender])
		Telepathy:RefreshForeignPanel(LinkedMembers[sender])
	end
end


--request types are "r [refresh] or a [ability]"
function Telepathy:ComRequest(requesttype, id, type, who)
	local msg = strconcat(requesttype,"|",id,"|",type)
	Telepathy:SendCommMessage(comRequest, msg, "WHISPER", who)
end

function Telepathy:OnComRequest(msg, channel, sender)
	if sender == UnitName("player") then
		if not teleDebug then return end
	end

	local req, tid, type = strsplit("|", msg)
	local id = tonumber(tid)
	Telepathy:Trace("OnComRequest(",req,id,type,")")

	local spellname = GetSpellInfo(id)
	
	Telepathy:Trace("Got ability request for",spellname)

	if req == "a" then
		--ask for an ability to be cast, highlight the button
		local info = PublishedAbilities[id]
		if info and info.button and info.highlight then
			local h = info.highlight
			Telepathy:Trace("Highlighting button")
			h:SetTexture(1, 1, 0, 0.5)
			h:Show()
			h.HideTime = GetTime() + 5
			if db.RequestSound and string.len(db.RequestSound) then
				PlaySoundFile(sounds[db.RequestSound])
			end
		else
			Telepathy:Trace("Not highlighting button")
			if info and not info.button then
				Telepathy:Trace("no button!")
			elseif info and not info.highlight then
				Telepathy:Trace("no highlight!")
			elseif not info then
				Telepathy:Trace("no info!")
			end
		end
	elseif req == "r" then
		--player is asking for my list of abilities
		Telepathy:Trace("Responding to refresh request...")
		Telepathy:ComRefresh("WHISPER", sender)
	end
end

function Telepathy:OnCooldownHighlightCheckTimer()
	local now = GetTime()
	
	for k,info in pairs(PublishedAbilities) do	
		local h = info.highlight
		if h and h:IsVisible() and h.HideTime then
			if now > h.HideTime then
				Telepathy:Trace("Hiding highlight texture")
				h:Hide()
				h.HideTime = 0
			end
		elseif not h then
			Telepathy:Trace("highlight for", v.Name, "spell id", info.id, "is nil")
		end	
	end
end

function Telepathy:CreateCombatFrames()
	
	if self.MainFrame then
		self.MainFrame:Hide()
		self.MainPanel:Hide()
		self.ForeignPanel:Hide()
		self.MainFrame = nil
		self.MainPanel = nil
		self.ForeignPanel = nil
	end

	self:CreateMainFrame()
	self:CreateMainPanel()
	self:CreateForeignPanel();
end

function Telepathy:OnEnable()
    --set up the config options
	self.db = LibStub("AceDB-3.0"):New("TelepathyDB", defaults, true)
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Telepathy", options, {"telepathy"})
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Telepathy", "Telepathy")

	db = self.db.char
	
	--load local values
	buttonPadding = db.ButtonPadding
	opacityCD = db.OpacityCD
	opacityAvailable = db.OpacityAvailable
	
	--create frames
	self:CreateCombatFrames()

	--register slash commands
	self:CreateSlashCommands()

	--register events
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")

	--register communication functions
	self:RegisterComm(comPublish, self.OnComPublish)
	self:RegisterComm(comCD, self.OnComNotifyCD)
	self:RegisterComm(comRequest, self.OnComRequest)
	self:RegisterComm(comRefresh, self.OnComRefresh)

	--load the previously published abilities, saved in the db.Published table
	for k,v in pairs(db.Published) do
		self:PublishAbility(v.id, false, true)
	end
	local c = GetCurrentChannel()
	self:ComRefresh(c)

	db.Published = PublishedAbilities

	--start cooldown highlight timer
	self:ScheduleRepeatingTimer("OnCooldownHighlightCheckTimer", 2, nil)

	--welcome message
	self:Print([[|cff20D6E6Welcome to the Telepathy beta!|r]])
	self:Print([[|cff20D6E6Open the Telepathy spellbook by typing|r |cffD9DC2A/telepathy|r]])
	
end

function Telepathy:OnDisable()
    -- Called when the addon is disabled
end
